;AV5U.ASM  ;********************************************************
;USB specific code for ATmega32U4 (essentially valid for ATmega32U2)
;*******************************************************************/

ISRGEN:  ;***********************************************************
;       USB GENeral Interrupt Service Routine (reacts only on End Of ReSeT)
;********************************************************************
	in SAVE_SREG,SREG
	lds INTREG,UDINT       	;if not End Of ReSeT Interrupt, then exit
    sbrs INTREG,EORSTI
    rjmp GEI10				;ignore any else
    lds	INTREG,UDINT		;most times Wakeup Interrupt Flag is set too
	andi INTREG,0xE7        ;clear End Of ReSeT Interrupt Flag and Wakeup Flag
	sts	UDINT,INTREG

;Initialize EP0
	clr INTREG			  	;select endpoint by number
    sts UENUM,INTREG
	lds	INTREG,UECONX	
    sbr INTREG,1<<EPEN		;set UECONX,EPEN (Bit 0) = activate EP
	sts	UECONX,INTREG
	ldi INTREG,EP0_CFG0
    sts UECFG0X,INTREG
 	ldi INTREG,EP0_CFG1
	sbr	INTREG,1<<ALLOC		;set ALLOC bit	
	sts	UECFG1X,INTREG

	lds	INTREG,UEIENX		
	sbr	INTREG,1<<RXSTPE    ;enable Endpoint Interrupts when a new
	sts	UEIENX,INTREG		;valid SETUP packet is received (RXSTPI=1)

GEI10:
    out SREG,SAVE_SREG		;restore SREG to MAIN level
    reti


ISRCOM:  ;***********************************************************
;       USB EndPoint Communication Interrupt Service Routine
;********************************************************************
	in SAVE_SREG,SREG	  
	push YL					;save for general use during this interrupt										
	push YH					;in the subrputines called from here, too
    lds INTREG,UEINT       	;EP number is given by bit position in UEINT
	sbrs INTREG,0			;EP0?
	rjmp COI10
	ldi	INTREG,0			;Select EP0
	sts UENUM,INTREG
    lds INTREG,UEINTX  
	sbrc INTREG,RXSTPI		;if no valid SETUP Paket, then exit
    rcall SETUP				;subroutine SETUP handles enumeration
COI10:

    lds INTREG,UEINT       	;in this template, EP1 is handled by interrupt
	sbrc INTREG,1			;sendCmdByte fills EP1_BUF
	rcall EP1_IN			
							
							
    lds INTREG,UEINT       	;EP number is given by bit position in UEINT
	sbrc INTREG,2			;EP2?
    rcall EP2_OUT 

;EP3 never responds to notification request 
;but must be present formally
    lds INTREG,UEINT       	
	sbrc INTREG,3			;EP3?
    rcall EP3_IN
	pop YH
	pop YL
	out SREG,SAVE_SREG		;restore SREG to MAIN level
    reti


SETUP:   ;***********************************************************
;UP SETUP: Handle a SETUP request (packet) from host (EP0)
;may only be called from USB EndPoint COMmunication Interrupt Service Routine
;********************************************************************
	push ZL					;save registers which are used at main level
	push ZH
;read 8 bytes of the SETUP packet into SRAM
    lds	INTREG,UEDATX     
	sts	SU_R_T,INTREG		;bmRequestType

	lds	INTREG,UEDATX
	sts	SU_REQ,INTREG		;bRequest

	lds	INTREG,UEDATX
	sts	SU_VAL,INTREG		;wValue

	lds	INTREG,UEDATX
	sts	SU_VAL+1,INTREG

	lds	INTREG,UEDATX
	sts	SU_IND,INTREG		;wIndexI

	lds INTREG,UEDATX
	sts	SU_IND+1,INTREG

	lds	INTREG,UEDATX
	sts	SU_LEN,INTREG		;wLength

	lds	INTREG,UEDATX
	sts	SU_LEN+1,INTREG
	lds	INTREG,UEINTX		;clear UEINTX,RXSTPI (Bit 3)		
	cbr INTREG,1<<RXSTPI	;ACK SU packet (terminate ISR and clear FIFO)
	sts	UEINTX,INTREG
	lds	INTREG,SU_R_T		;check if class specific request
	cpi INTREG,0x21			
	brne SET10
	rjmp SET65
SET10:
	cpi INTREG,0xA1
	brne SET12
	rjmp SET60

SET12:
	andi INTREG,0x61		;or a Standard Device Request 
	breq SET15	    		;(exclusively device or interface; go on with SET15)    	
	cpi INTREG,1
	breq SET15
	rjmp SET18

SET15:	
	lds	INTREG,SU_REQ		;check Request type
	cpi	INTREG,0x00    		;Get_Status?
	breq SET20
	cpi	INTREG,0x05    		;Set_Adress?
	breq SET25		
	cpi	INTREG,0x06    		;Get_Descriptor?
	breq SET30
	cpi	INTREG,0x09    		;Set_Configuration?
	brne SET18				;if not, then exit
	rjmp SET80
SET18:	
	rjmp SET92				;unsupported Req. -> STALL and exit

SET20:
;Get_Status
	clr INTREG
	sts	UEDATX,INTREG
	sts	UEDATX,INTREG
	lds	INTREG,UEINTX		;clear UEINTX,TXINI (Bit 0)
	andi INTREG,0xFE		;send packet
	sts	UEINTX,INTREG
SET22:	
	lds	INTREG,UEINTX		;wait until RXOUTI is set, ZLP (ACK) from PC
	sbrs INTREG,RXOUTI
	rjmp SET22		
	rjmp SET95				;exit

SET25: 
;Set_Adress
	lds	INTREG,SU_VAL		;LSB(wValue) of Set_Adress contains device address Geraeteadr.
	andi INTREG,0x7F			;0-127
	sts	UDADDR,INTREG		;save address
	lds	INTREG,UEINTX		;clear UEINTX,TXINI (Bit 0)		
	cbr INTREG,1<<TXINI		;send ZLP packet (signals success, clears EP bank)
	sts	UEINTX,INTREG
SET28:	
	lds	INTREG,UEINTX		;wait until TXINI set again
	sbrs INTREG,TXINI		;ACK fromPC, EP bank clear
	rjmp SET28
	lds	INTREG,UDADDR					
	sbr INTREG,1<<ADDEN		;set UDADDR,ADDEN (Bit 7) -- activate new device address
	sts	UDADDR,INTREG
	rjmp SET95				;exit
	
SET30:	
;Get_Descriptor
	lds	INTREG,SU_VAL+1		;HIGH(Su_Val) contains Deskriptor type/number
	cpi	INTREG,1			;Device Deskriptor
	breq SET32
    cpi	INTREG,2			;Configuration Descriptor
	breq SET38
	cpi	INTREG,3			;String Descriptor
	breq SET50
	rjmp SET92				;unsupported -->Stall??

SET32: 
;Device Descriptor
    ldi ZL,LOW(CDC_DEVICEDESCRIPTOR*2)
	ldi	ZH,HIGH(CDC_DEVICEDESCRIPTOR*2)
	lpm YH,Z
	rcall sendDeviceDescriptor
	rjmp SET95

SET38: 
;Configuration Descriptor
    lds	YH,SU_LEN			;how many bytes requested?
    cpi YH,0           		;accept min.1, max.255 bytes
    breq SET40
    lds INTREG,SU_LEN+1    	
    cpi INTREG,0           	
    brne SET40
    cpi YH,67       		;limit to length of actual CDC_ACM class descriptor
	brcs SET42
SET40: 
	ldi YH,67				;CDC_ACM
SET42: 
							;slower USB cycle for better DMX stability
	ldi	ZL,LOW(CDC_SLOW_CONFIGDESCRIPTOR*2)
	ldi	ZH,HIGH(CDC_SLOW_CONFIGDESCRIPTOR*2)
	rjmp SET58	

SET50:     
	lds	INTREG,SU_VAL		;contains string index number
	tst INTREG
	brne SET52
	ldi	ZL,LOW(MANUFACT_STRINGDESCRIPTOR*2)
	ldi	ZH,HIGH(MANUFACT_STRINGDESCRIPTOR*2)
	rjmp SET55
SET52:
	cpi	INTREG,1			;PRODUCT String Descriptor (is always =1 in this template) 
	brne SET54
	ldi	ZL,LOW(PRODUCT_STRINGDESCRIPTOR*2)
	ldi	ZH,HIGH(PRODUCT_STRINGDESCRIPTOR*2)
	rjmp SET55
SET54:	
	rjmp SET92

SET55:	
	lpm	YH,Z
SET58: 
	rcall sendDescriptor
	rjmp SET95					;exit

;Code until label SET80 only used when CDC_CLASS is active:
SET60: 
	lds	INTREG,SU_REQ			;check type of class request
	cpi	INTREG,0x21        		;GET_LINE_CODING
	breq SET62
	rjmp SET92					;not supported
SET62:
    ldi ZL,LOW(FAKELINECODING*2)
	ldi	ZH,HIGH(FAKELINECODING*2)
	ldi	YH,7					;number of bytes
	rcall sendDescriptor
	rjmp SET95

SET65: 
	lds	INTREG,SU_REQ			;check type of class request
	cpi	INTREG,0x20        		;SET_LINE_CODING
	brne SET70
SET68:
	lds	INTREG,UEINTX			;wait until RXOUTI set again
	sbrs INTREG,RXOUTI			;(ZLP (success) from PC)
	rjmp SET68
	lds	INTREG,UEINTX					
	cbr INTREG,1<<RXOUTI		;if yes then confirm by ACK handshake 
	sts	UEINTX,INTREG  
	rjmp SET72

SET70:
	cpi	INTREG,0x22        		;SET_CONTROL_LINE_STATE
	breq SET72
	rjmp SET92					;unsupported Req. -> STALL and exit
SET72:
	lds	INTREG,UEINTX			;clear UEINTX,TXINI (Bit 0)FE			
	cbr INTREG,1<<TXINI			;send ZLP packet (signals success, clears EP bank)
	sts	UEINTX,INTREG
SET75:
	lds	INTREG,UEINTX			;wait until TXINI set again 
	sbrs INTREG,TXINI			;signals ACK from PC, EP Bank clear again
	rjmp SET75
	rjmp SET95					;exit

SET80:	
;Set_Configuration
;all used EPs have to be disabled first (backwards, see manual p.261)
	ldi	YL,3					;CDC_ASC: init. counter for 3 endpoints
						
SET82:	
	sts	UENUM,YL				;first address highest endpoint
	lds	INTREG,UECONX		
	cbr INTREG,1<<EPEN			;switch EP OFF (bit0)
	sts	UECONX,INTREG
	lds	INTREG,UECFG1X		
	cbr INTREG,1<<ALLOC			;release DRAM bank (bit1)
	sts	UECFG1X,INTREG
	dec	YL
	brne SET82

;Init EP1 (IN EP: set parameters and enable NAKIN interrupt)
	ldi INTREG,1		   		;select endpoint access by number
    sts UENUM,INTREG
	lds	INTREG,UECONX	
    sbr INTREG,1<<EPEN			;set UECONX,EPEN (Bit 0) = activiate EP
	sts	UECONX,INTREG
 	ldi INTREG,EP1_CFG0
    sts UECFG0X,INTREG
	ldi INTREG,EP1_CFG1
	sbr	INTREG,1<<ALLOC			;set ALLOC bit	
	sts	UECFG1X,INTREG

    lds INTREG,UEIENX      	
 	sbr INTREG,1<<TXINE
    sts UEIENX,INTREG      		;(interrupt is released by HW
	 							;when host IN request is responded with NAK)
							
;Init EP2 (OUT EP: set parameters and enable RXOUT interrupt
	ldi INTREG,2   				;select endpoint access by number
    sts UENUM,INTREG
	lds	INTREG,UECONX		
    sbr INTREG,1<<EPEN			;set UECONX,EPEN (Bit 0) = activiate EP
	sts	UECONX,INTREG
 	ldi INTREG,EP2_CFG0
    sts UECFG0X,INTREG
	ldi INTREG,EP2_CFG1
	sbr	INTREG,1<<ALLOC			;set ALLOC bit	
	sts	UECFG1X,INTREG

    lds INTREG,UEIENX      		
    sbr INTREG,1<<RXOUTE|1<<NAKINE ;set UEIENX,RXOUTE (Bit 2) = enable RX OUT (EP2) interrupt
    sts UEIENX,INTREG      	;(interrupt is released by HW when new OUT data available)

;Init EP3 (IN EP: set parameters and enable NAKIN interrupt)
	ldi INTREG,3		   		;select endpoint access by number
    sts UENUM,INTREG
	lds	INTREG,UECONX		
    sbr INTREG,1<<EPEN			;set UECONX,EPEN (Bit 0) = activiate EP
	sts	UECONX,INTREG
 	ldi INTREG,EP3_CFG0
    sts UECFG0X,INTREG
	ldi INTREG,EP3_CFG1
	sbr	INTREG,1<<ALLOC			;set ALLOC bit	
	sts	UECFG1X,INTREG

    lds INTREG,UEIENX      	
 	sbr INTREG,1<<TXINE
    sts UEIENX,INTREG      		;(interrupt is released by HW
	 							;when host IN request is responded with NAK)
                        
SET85:
	clr	INTREG					;select EP0 
	sts	UENUM,INTREG
	lds	INTREG,UEINTX		
	cbr INTREG,1<<TXINI			;send ZLP packet (signals succes, clears EP bank)
	sts	UEINTX,INTREG
SET88:	
	lds	INTREG,UEINTX			;wait until TXINI is set agian -->ready
	andi INTREG,1<<TXINI		;ACK from PC, EP Bank free again)
	breq SET88
	sbr AVIFLAGS,m_USBCONNECTED  
	rjmp SET95
SET92: 
	lds	INTREG,UECONX			;set UECONX,STALLRQ	(Bit 5)				
	sbr INTREG,1<<STALLRQ		;no Standard Request! ask for STALL answer
	sts	UECONX,INTREG
SET95: 
	pop	ZH
	pop	ZL
    ret


sendDeviceDescriptor:  ;*********************************************
;send Device Descriptor to host
;parameter YH
;YH,YL modified, transferred over to "sendDescriptor"label SND25
;********************************************************************
	lds INTREG,SU_LEN+1
	tst INTREG
	brne SDD20
	lds INTREG,SU_LEN
	cp INTREG,YH				;YH is length declared in descriptor
	brcc SDD10
	mov YH,INTREG				;limit length to reqest
SDD10:
	clr YL						;set number of written bytes to zero
SDD20: 
	lds	INTREG,UEINTX			;interruption from PC (Host) then exit
	sbrc INTREG,RXOUTI
	rjmp SND80
;send packet after each 8 bytes and clear FIFO bank
	lpm	INTREG,Z+				;byte into FIFO
	sts	UEDATX,INTREG
	dec	YH						;decrease counter
	inc	YL
	cpi YL,8
	brne SDD20
	lds	INTREG,UEINTX			;clear UEINTX,TXINI (Bit 0)
	andi INTREG,0xFE			;send packet
	sts	UEINTX,INTREG
SDD30:	
	lds	INTREG,UEINTX			;wait until TXINI is set again (then is EP bank free agin) 
	andi INTREG,0x05			;or until RXOUTI is set (ACK (ZLP) from PC)
	breq SDD30

	lds	INTREG,UEINTX			;interruption from PC (Host) then exit
	sbrc INTREG,RXOUTI
	rjmp SND80
;now insert 4 bytes VID/PID from EEPROM or defazult values
	push ZL
	push ZH
	push r24
	push r25
	push TEMP1
;intermediately, read 4 bytes from EEPROM or insert default
;everything here handled in little endian
	rcall getVID
	sts UEDATX,r24
	sts UEDATX,r25

	rcall getPID	
	sts UEDATX,r24
	sts UEDATX,r25

	pop TEMP1	
	pop r25
	pop r24
	pop ZH
	pop ZL
	ldi YL,12					;bytes sent
	subi YH,4
	rjmp SND25					;continue with standard routine


getVID:   ;**********************************************************
;destroys TEMP1,ZL,ZL
;return 24=lowByte, r25=HighByte
;********************************************************************
	ldi ZH,HIGH(VENDORIDL)
	ldi ZL,LOW(VENDORIDL)
	call comReadEep
	mov r24,r0
	call comReadEep
	mov r25,r0
	mov ZL,r25
	or ZL,r24
	brne GVD10

	ldi r24,VIDDEFL
	ldi r25,VIDDEFH
GVD10:
	ret


getPID:   ;**********************************************************
;destroys TEMP1,ZL,ZL
;return 24=lowByte, r25=HighByte
;********************************************************************
	ldi ZH,HIGH(PRODUCTIDL)
	ldi ZL,LOW(PRODUCTIDL)
	call comReadEep
	mov r24,r0
	call comReadEep
	mov r25,r0
	mov ZL,r25
	or ZL,r24
	brne GPD10

	ldi r24,PIDDEFL
	ldi r25,PIDDEFH
GPD10:
	ret


comReadEep:   ;******************************************************
;ZL,ZH in EEARL,H schreiben und EERE (=Bit0) in EECR setzen
;return Einerkompelement(r0)
;anschliessend Z inkrementieren
;********************************************************************
	call addressIntEep
	in r0,EEDR
	com r0
	adiw ZL,1
	ret


sendDescriptor:  ;***************************************************
;send Descriptor to host
;YH transferred from calling routine
;ZL/ZH parameter, modified, but not used in calling routine anymore
;YH/YL parameter at jumpIn from sendDeviceDescriptor
;YH/YL modified
;********************************************************************
	lds INTREG,SU_LEN+1
	tst INTREG
	brne SND10
	lds INTREG,SU_LEN
	cp INTREG,YH
	brcc SND10
	mov YH,INTREG				;limit length to reqest
SND10:
	clr YL						;set number of written bytes to zero

SND20: 
	lds	INTREG,UEINTX			;interruption from PC (Host) then exit
	sbrc INTREG,RXOUTI
	rjmp SND80
	wdr
;send packet after each 8 bytes and clear FIFO bank
	lpm	INTREG,Z+				;byte into FIFO
	sts	UEDATX,INTREG
	dec	YH						;decrease counter
	inc	YL

SND25:
;entry point from sendDeviceDescriptor
	ldi	INTREG,7				;check if FIFO is filled (8 bytes in this case)
	and	INTREG,YL				;Modulo by AND with 7 ls bits
	brne SND40
	lds	INTREG,UEINTX			;clear UEINTX,TXINI (Bit 0)
	andi INTREG,0xFE			;send packet
	sts	UEINTX,INTREG
SND30:	
	lds	INTREG,UEINTX			;wait until TXINI is set again (then is EP bank free agin) 
	andi INTREG,0x05			;or until RXOUTI is set (ACK (ZLP) from PC)
	breq SND30
SND40:	
	cpi	YH,0
	brne SND20
;check if ACK(ZLP) received from PC (then ready)
;(happens when a multiple of FIFO buffer n*8 was requested)
	lds	INTREG,UEINTX			;check if ACK (ZLP) from PC arrived
	andi INTREG,0x04  			;(RXOUTI flag set?)
	breq SND50
	rjmp SND70					;send ACK and exit
SND50: 
	lds	INTREG,UEINTX					
	cbr INTREG,1<<TXINI			;send remaing partial packet(clears EP bank)
	sts	UEINTX,INTREG

SND60:	
	wdr
	lds	INTREG,UEINTX			;wait until RXOUTI is set again
	sbrs INTREG,RXOUTI			;(ZLP (success) from PC)
	rjmp SND60
SND70:	
	lds	INTREG,UEINTX					
	cbr INTREG,1<<RXOUTI		;if yes then confirm by ACK handshake
	sts	UEINTX,INTREG
SND80:	
	ret
	

EP1_IN:   ;**********************************************************
;host PC requests notification from this device
;********************************************************************
	push TEMP1             	;save registers which are used at main level

	ldi	INTREG,1			;select EP1
	sts UENUM,INTREG

    lds INTREG,UEINTX      	;check if data were requested by host PC
	sbrs INTREG,TXINI		;Bit0
	rjmp EP180

	push YL
	push YH

	andi INTREG,0xA0		;clear all flags except FIFOCON and RWAL
    sts UEINTX,INTREG
    sbrs YFLAGS,EP1TRIGGER	;data available ?
    rjmp EP160
    ldi YH,HIGH(EP1_BUF)
	lds INTREG,EP1_HEAD
	lds YL,EP1_TAIL
EP120:
	ld TEMP1,Y+          	;copy byte from SRAM buffer into EP1 FIFO
    sts UEDATX,TEMP1 
	cpi YH,HIGH(EP1_BUFTOP)
	brne EP130
    ldi YH,HIGH(EP1_BUF)
EP130:
    cp YL,INTREG
    brne EP120
	sts EP1_TAIL,YL
	cbr YFLAGS,m_EP1TRIGGER	; - clear flag
EP160: 
   	lds INTREG,UEINTX      	;(Bit 7, FIFO Control Bit)
    cbr INTREG,1<<FIFOCON   ;clearing of FIFO Control Bit releases data transmission
    sts UEINTX,INTREG      	
	pop YH
	pop YL

EP180: 
    pop TEMP1
    ret


EP2_OUT:  ;**********************************************************
;this device receives data from host PC
;modifies YH/YL without effect on calling routine
;********************************************************************
	push r24
	push TEMP1
	push TEMP2
	push TEMP3
	ldi	INTREG,2				;select EP2
	sts UENUM,INTREG
	lds INTREG,UEINTX      		;check if new data from PC available
    sbrs INTREG,RXOUTI     	
    rjmp EP280          		;if RXOUTI flag (bit2) not set then exit

	lds INTREG,UEINTX    	
	cbr INTREG,1<<RXOUTI		;RXOUTI interrupt is kept alive while EP2 is locked
	sts UEINTX,INTREG			;else system may hang up or get extremely slow

	lds INTREG,UEBCLX			;number of new bytes in the FIFO
	tst INTREG
	brne EP210
	rjmp EP270				

EP210: 
;INTREG is loop counter
	wdr
    lds TEMP1,UEDATX      		;copy byte into SRAM

	sbrc AVIFLAGS,COMPILE
	rjmp EP240
	sbrc XFLAGS,NOBACKGND		;if set,ignore low level background messages
	rjmp EP240
	cpi TEMP1,ESC				;ESC=brutal termination of endless loops etc
	brne EP215
	rjmp EP280					;resets RETURN stack, performs ABOORT
EP215:
;byte received from background message without storage in buffer
	cpi TEMP1,CTRL_S      		;immediately send .S stack report
	brne EP225
	sbr YFLAGS,m_BAKDOTS		;executed in timer0 interrupt handler
	rjmp EP265
EP225:
	sbrc YFLAGS,GETVABC			;if set:executed in timer0 interrupt handler
	rjmp EP240
	cpi TEMP1,CTRL_A      	
	brne EP230
	sbr YFLAGS,m_GETVABC | m_SAVEVA
	cbr YFLAGS,m_SAVEVB
	rjmp EP265
EP230:
	cpi TEMP1,CTRL_B      	
	brne EP235
	sbr YFLAGS,m_GETVABC | m_SAVEVB
	cbr YFLAGS,m_SAVEVA
	rjmp EP265
EP235:
	cpi TEMP1,CTRL_C
	brne EP240
	sbr YFLAGS,m_GETVABC | m_SAVEVA | m_SAVEVB
	rjmp EP265

EP240:
;normal buffer handling
	lds YH,USBRXHEADH
   	lds YL,USBRXHEADL	 		;init pointer to SRAM buffer
    st Y+,TEMP1
	cpi YH,HIGH(EP2_BUFTOP)		;buffer top msut be at page boundary	
	brne EP245
	ldi YH,HIGH(EP2_BUF)
	ldi YL,LOW(EP2_BUF)
EP245:
	sts USBRXHEADH,YH
	sts USBRXHEADL,YL
	lds TEMP2,USBRXTAILH
	lds TEMP3,USBRXTAILL

;calculate fill level of EP2_BUF
;TAIL>HEAD?
	cp YH,TEMP2					;cp USBRXHEADH,USBRXTAILH
	brcs EP255					;HEAD truly <TAIL, transform to absolute value
	breq EP250
	rjmp EP260					;HEAD truly > TAIL, standardd case
EP250:
;High Bytes equal, then compare Low Bytes
	cp YL,TEMP3					;cp USBRXHEADL,USBRXTAILL	
	brcs EP255
	breq EP265					;HEAD=TAIL, no action needed
	rjmp EP260	
EP255:
;Absolut value of cyclic buffer: ad buffer size to HEAD
;because buffer designed to be always full pages:
	inc YH
	inc YH
	inc YH						;add buffer size number of pages
EP260:
;buffer fill level = HEAD-TAIL
	sub YL,TEMP3				;USBRXTAILL
	sbc YH,TEMP2				;USBRXTAILH
	cpi YH,(EP2_SIZE-1)
	brcs EP265
;max. 256 bytes freely available in buffer
;detailled compare with LowByte	
	cpi YL,0x70			;at least 128bytes must be reserved for 2 FIFO packets
	brcs EP265
;lock EP2 FIFOCON interrupt when EP2_BUF contains at least (SIZE-64) bytes
	sbr YFLAGS,m_EP2LOCK ;gets unlocked in routine pullUSBRXBuf 
						;as soon as EP2_BUF contains less than (SIZE-64) bytes
						;this way EP2_BUF always can accept
						;another packet of 64 bytes without data loss
						;(similar to CTS/RTS handshake at RS-232)

EP265:
	dec INTREG
	breq EP270
    rjmp EP210

EP270:
	sbrc YFLAGS,EP2LOCK
	rjmp EP275
	lds INTREG,UEINTX    	
	cbr INTREG,1<<RXOUTI
	sts UEINTX,INTREG			;insert a small delay 
	nop							;because RXOUTI shall be cleared before FIFOCON	
	lds INTREG,UEINTX		
   	cbr INTREG,1<<FIFOCON   	;clear UEINTX,FIFOCON (Bit 7, FIFO Control Bit)    
   	sts UEINTX,INTREG			;message that FIFO bank is free again
EP275:
	pop TEMP3
	pop TEMP2
	pop TEMP1
	pop r24
    ret
		
EP280:

	lds INTREG,UEINTX    	
	cbr INTREG,1<<RXOUTI
	sts UEINTX,INTREG			;insert a small delay 
	nop							;because RXOUTI shall be cleared before FIFOCON	
	lds INTREG,UEINTX		
   	cbr INTREG,1<<FIFOCON   	;clear UEINTX,FIFOCON (Bit 7, FIFO Control Bit)    
   	sts UEINTX,INTREG			;message that FIFO bank is free again

	ldi INTREG,LOW(RETURNSTACK-2);readjust return stack
	out SPL,INTREG
	ldi INTREG,HIGH(RETURNSTACK-2)
	out SPH,INTREG
	sei
	jmp ABORT					;shall ret to MAIN level


EP3_IN:   ;**********************************************************
;host PC requests notification from this device
;modifies YH/YL without effect on calling routine
;********************************************************************
	push TEMP1             	;save registers which are used at main level

	ldi	INTREG,3			;select EP3
	sts UENUM,INTREG

    lds INTREG,UEINTX      ;check if data were requested by host PC
	sbrs INTREG,TXINI		;Bit0
	rjmp EP380

	andi INTREG,0xA0		;clear all flags except FIFOCON and RWAL
    sts UEINTX,INTREG
   
    sbrs YFLAGS,EP3TRIGGER    ;data available ?
    rjmp EP335

    ldi YL,LOW(EP3_BUF) 	;init pointer to SRAM buffer
    ldi YH,HIGH(EP3_BUF)
    lds INTREG,EP3_CNT     	;loop counter = EP3_CNT
EP320: 
	ld TEMP1,Y+          	;copy byte from SRAM buffer into EP3 FIFO
    sts UEDATX,TEMP1     
    dec INTREG
    brne EP320

EP335: 
	cbr YFLAGS,m_EP3TRIGGER	; - clear flag

    lds INTREG,UEINTX      	;(Bit 7, FIFO Control Bit)
    cbr INTREG,1<<FIFOCON   ;clearing of FIFO Control Bit releases data transmission
    sts UEINTX,INTREG      	

EP380: 
    pop TEMP1
    ret

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

;helpful for debugging when USB is critical
/*
sendDebugByte:  ;****************************************************
;Parameter=r24, not destructive
;********************************************************************
	wdr
	push r25			
SAB10:
	lds r25,UCSR1A
	sbrs r25,UDRE1
	rjmp SAB10
	sts UDR1,r24
	pop r25
	ret


sendDebugASCII:   ;**************************************************
;send hex representation of one byte in ASCII text format
;Parameter=r24, not destructive
;********************************************************************
	push r24
	ldi r24,0x20
	rcall sendDebugByte
SDBA1:   
	pop r24
	push r24                ;copy stack
	swap r24
	andi r24,0x0F
	subi r24,0xD0           ;add 0x30 
	cpi r24,0x3A
	brcs SDBA2
	subi r24,0xF9           ;add 7
SDBA2:   
	cpi r24,0x30
	brne SDBA4
	ldi r24,0x20             ;suppress leading zero
SDBA4:   
	rcall sendDebugByte      ;transmit high nibble first
SDBA6:   
	pop r24
	push r24                ;copy stack
	andi r24,0x0F
	subi r24,0xD0           ;add 0x3
	cpi r24,0x3A
	brcs SDBA8
	subi r24,0xF9           ;add 7
SDBA8:   
	rcall sendDebugByte	    ;transmit low nibble last
	pop r24
	ret
*/
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

initUSB:  ;**********************************************************
;enable USB interface (Full-Speed 12Mbit/s) and related interruptsi
;********************************************************************
	clr r0
    ldi r25,0x01        ;UHWCON = 0x01, enable Device Modus
    sts UHWCON,r25      ; and power supply for D+ and D- lines
;if CPU==32U2: REGCR with inverted polarity instead
	ldi r25,1<<USBE|1<<FRZCLK|1<<OTGPADE      
						;= 0xB0 , enable USB Rx/Tx, clock and VBUS line
   	sts USBCON,r25      

    nop   
    cbr r25,1<<FRZCLK   ;FRZCLK=0, necessary, else no WAKEUP Interrupt
    sts USBCON,r25
	nop
    sbr r25,1<<FRZCLK   ;FRZCLK=1, save current
    sts USBCON,r25
        
;start PLL
    ldi r25,PLLPRE      ;PLL prescaler for crystal
						;ATmega32U2:confusing difference between manual and def.inc
						;here bit2 (def.inc bit PLLP0 corresponds with manual PINDIV)
    out PLLCSR,r25
    ori r25,0x02        ;PLLE (PLL Enable Bit) = 1, start PLL
    out PLLCSR,r25
IUS10: 
	in r25,PLLCSR      ;wait until PLOCK = 1
    sbrs r25,PLOCK
    rjmp IUS10

    lds r25,USBCON      ;FRZCLK=0: enable clock
    cbr r25,1<<FRZCLK
    sts USBCON,r25

    clr r25             ;UDCON = 0 attach the device
    sts UDCON,r25
    ldi r25,1<<EORSTE
    sts UDIEN,r25       ;enable End Of ReSeT Interrupt

;init USB receive buffer
 	ldi r25,HIGH(EP2_BUF)
	sts USBRXHEADH,r25
	sts USBRXTAILH,r25
 	ldi r25,LOW(EP2_BUF)
	sts USBRXHEADL,r25
	sts USBRXTAILL,r25
    sei                 ;enable interrupts globally
    ret


setVid:  ;(hexWord -- ) ;********************************************
;********************************************************************
	mov TEMP1,TOSL
	mov TEMP2,TOSH
	push TEMP2
	push TEMP1				;bugfix 07Nov23
	ld TOSL,X+				;emulate DROP
	ld TOSH,X+
	
	push AVIFLAGS
	sbr AVIFLAGS,(m_TEMPHEX | m_NOLEADS)
	sbr XFLAGS,m_LEADZERO
	ldi r24,LOW(OLDSTR*2)
	call pageSendString
	ldi r24,LOW(VIDSTR*2)
	call pageSendString

	rcall getVID
	push r24
	ldi r24,'0'
	call sendCmdByte
	ldi r24,'x'
	rcall sendCmdByte
	mov r24,r25
	call sendASCII
	pop r24
	push r24
	call sendASCII
	pop r24
	ldi r24,LOW(NEWSTR*2)
	call pageSendString
	ldi r24,LOW(VIDSTR*2)
	call pageSendString
	ldi r24,'0'
	rcall sendCmdByte
	ldi r24,'x'
	rcall sendCmdByte
	mov r24,TEMP2
	call sendASCII
	mov r24,TEMP1
	call sendASCII
	cbr XFLAGS,m_LEADZERO

	ldi r24,LOW(WORDSTR*2)
	call pageSendString
	pop AVIFLAGS

	call KEY
	cpi TOSL,'Y'
	brne ASV90
	ld TOSL,X+				;emulate DROP
	ld TOSH,X+

	pop TEMP1
	pop TEMP2		
	ldi ZL,VENDORIDL
	clr ZH
	com TEMP1
	call burnIntEep
	adiw ZL,1
	mov TEMP1,TEMP2
	com TEMP1
	call burnIntEep
	ret

ASV90:
	pop TEMP1
	pop TEMP2
	jmp parameterErr


setPid:   ;(hexWord -- ) ********************************************
;********************************************************************
	mov TEMP1,TOSL
	mov TEMP2,TOSH
	push TEMP2
	push TEMP1				;bugfix 07Nov23
	ld TOSL,X+				;emulate DROP
	ld TOSH,X+

	push AVIFLAGS
	sbr AVIFLAGS,(m_TEMPHEX | m_NOLEADS)
	sbr XFLAGS,m_LEADZERO
	ldi r24,LOW(OLDSTR*2)
	call pageSendString
	ldi r24,LOW(PIDSTR*2)
	call pageSendString

	rcall getPID
	push r24
	ldi r24,'0'
	rcall sendCmdByte
	ldi r24,'x'
	rcall sendCmdByte
	mov r24,r25
	rcall sendASCII
	pop r24
	push r24
	rcall sendASCII
	pop r24
	ldi r24,LOW(NEWSTR*2)
	call pageSendString
	ldi r24,LOW(PIDSTR*2)
	call pageSendString
	ldi r24,'0'
	rcall sendCmdByte
	ldi r24,'x'
	rcall sendCmdByte
	mov r24,TEMP2
	rcall sendASCII
	mov r24,TEMP1
	rcall sendASCII
	cbr XFLAGS,m_LEADZERO

	ldi r24,LOW(WORDSTR*2)
	call pageSendString
	pop AVIFLAGS
	
	call KEY
	cpi TOSL,'Y'
	brne ASV90
	ld TOSL,X+				;emulate DROP
	ld TOSH,X+
	
	pop TEMP1
	pop TEMP2
	ldi ZL,PRODUCTIDL
	clr ZH
	com TEMP1
	call burnIntEep
	adiw ZL,1
	mov TEMP1,TEMP2
	com TEMP1
	call burnIntEep
	ret

;********************************************************************
; Deskriptors in Program Flash (Word access = 2Byte  LSB first!)
;********************************************************************

CDC_DEVICEDESCRIPTOR:
;the assembler demands word formatted flash entries, so parts cannot be separated exactly
;Device Descriptor
.DB		18,1         	;bLength			length of this descriptor=18 bytes
                    	;descriptor type 	DEVICE Descriptor = 1
;.DB 	0x00,0x02       ;bdcUSB				current rev of USB spec= 0x0110 = USB2.00 
						;(host asks for HighSpeed mode, request not supported here.) 
						;Therefore better take:
.DB 	0x10,0x01       ;bcdUSB				current rev of USB spec= 0x0110 = USB1.10
.DB 	2,0       		;bDeviceClass		COMMUNICATION DEVICE CLASS = 2
                    	;bDeviceSubClass 	= 0x00 unused
.DB 	0,8		    	;bDevice Protocol 	= 0x00, unused
                    	;bMaxPacketSize0	Endp.0 = 8
;until here 8 bytes

;in between 4 bytes from EEPROM

;some proposals:
;if your device shall appear as Arduino Micro
;.DB 	0x41,0x23       ;idVendor 			VID= 0x2341 = simulate Arduino
;.DB 	0x37,0x00       ;idProduct 			PID= 0x0037 = Micro
;if your device shall appear as experimental/prototype one
;.DB 	0x66,0x66       ;idVendor 			VID= 0x6666 for experimental use 
;.DB 	0xfc,0xaf       ;idProduct 			PID= 0xaffc
;insert 4 bytes from EEPROM and
;continue from there:
.DB     REVNUM,HW_VERSION;bcd_Device		release code
.DB 	0,1   			;iManufacturer		index of manufacturer string descriptor (unused)
                    	;iProduct			Index of product name string descriptor = 1
.DB 	0,1     		;iSerialNumber		serial number (unused)
                    	;bNumConfigurations	= 1 only one available
                        
CDC_CONFIGDESCRIPTOR:
;Configuration Descriptor
.DB     9,2          	;bLength			length of this descriptor = 9 bytes
                        ;bDescriptor type	CONFIGURATION descriptor = 2
.DB     67,0 			;wTotalLength		total length of configuration block = 67 bytes
                        
.DB     2,1      		;bNumInterfaces		two interfaces incl this descriptor
						;bConfigurationValue ID of this configuration = 1
#ifdef SELFPOWERED
.DB     0,0x40          ;iConfiguration		index of string descriptor:unused=0
						;bmAttributes		SELF-powered - no remote wakeup
.DB     0,9             ;MaxPower 			= 0 (self-powered only)
						;bLength (of next)	length of this descriptor = 9 bytes
#else
.DB     0,0x80          ;iConfiguration		index of string descriptor:unused=0
						;bmAttributes		BUS-powered - no remote wakeup
.DB     49,9            ;MaxPower			entry*2 = 98mA
						;bLength of next descriptor = 9 bytes
//9 belongs to next block.Put here due to HalfWord configuration
#endif
;COMMUNICATION DEVICE CLASS standard INTERFACE Descriptor  
;length of this descriptor = 9 bytes
.DB     4,0          	;bDescriptorType	INTERFACE descriptor type=4 
                        ;bInterfaceNumber	index of this interface =0 
.DB     0,1		        ;bAlternateSetting	index of this setting =0 
                        ;bNumEndpoints 		(Notification endpoint) = 1 
.DB     2,2       		;bInterfaceClass 	= 2 (COMMUNICATION)
                        ;bInterfaceSubClass = 2 (ABSTRACT_CONTROL_MODEL)
.DB     0,0         	;bInterfaceProtocol = 0 (unused)
                        ;iInterface 		= 0 (unused)
;COMMUNICATION DEVICE specific functional descriptor
.DB     5,0x24       	;bLength			length of this descriptor = 5 bytes
                        ;bDescriptorType	CS_INTERFACEdescriptor type = 0x24
.DB     0,0x10       	;bDescriptortSubtype HEADER subtype
.DB     1,5             ;bcdADC				=0x0100 revision of class specific descriptors
						;bLength of next descriptor = 5 bytes
;CALL MANAGEMENT functional Descriptor
;length of this descriptor = 5 bytes
.DB     0x24,1 			;bDescriptorType	CS_INTERFACEdescriptor type = 0x24
						;bDescriptorSubtype	
.DB     3,0   			;bmCapabilities		Device handles call management by itself 
						;					ia DataInterface
						;					i.e. in this case: no call management required
						;bDataInterface:	Data interface no.0 (see below) associated
;ABSTRACT CONTROL management functional Descriptor 
.DB     4,0x24         	;bLength			length of this descriptor = 4 bytes
						;bDesriptorType		CS_INTERFACEdescriptor type = 0x24
.DB     2,7				;bDescriptorSubtype
						;bmCapabilities		:Send_Break,Set_Line_Coding... etc
;UNION functional descriptor
.DB 	5,0x24          ;bLength			length of this descriptor = 5 bytes
                        ;bDescriptorType	CS_INTERFACE descriptor
.DB 	6,0		       	;bDescriptorSubtype UNION FUNCTIONAL descriptor
                        ;bMasterInterface
.DB     1,7  			;bSlaveInterface0    
						;bLength of next descriptor = 7 bytes    
;Notification standard endpoint descriptor
.DB		5,0x83			;bDescriptorType	ENDPOINT
						;bEndpointAddress	n0.3, input
.DB		3,16			;bmAttributes 		INTERRUPT = 3
						;wMaxPacketSize		= 16
.DB		0,64			;interval in ms		= 64
			            
;DATA CLASS INTERFACE descriptor
.DB		9,4				;bLength			length of this descriptor = 9 bytes
						;bDescriptorType	INTERFACE descriptor type
.DB		1,0				;bInterfaceNumber
						;bAlternateSetting
.DB		2,0x0A			;bNumEndpoints
						;bInterfaceClass	Data Class
.DB		0,0				;bInterfaceSubclass unused
						;bInterfaceProtocol no specific protocol
.DB		0,7				;iInterface 		unused
						;bLength of next descriptor = 7 bytes
;DATA BULK IN endpoint standard descriptor
.DB		5,0x81			;bDescriptorType	ENDPOINT
						;bEndpointAddress	nr.1, input 
.DB		2,64			;bmAttributes		= 2 (bulk)
						;wMaxPacketSize		= 64
.DB		0,1				;bInterval			=1ms

;DATA BULK OUT endpoint standard descriptor
.DB		7,5				;bLength			length of this descriptor = 7 bytes
						;bDescriptorType	ENDPOINT 
.DB		2,2				;bEndpointAddress	nr.=2, output
						;bmAttributes 		= 2 (bulk)
.DB		64,0			;wMaxPacketSize		= 64
.DB		1,0				;bInterval			=1ms
						;padding byte

						
CDC_SLOW_CONFIGDESCRIPTOR:
;Configuration Descriptor
.DB     9,2          	;bLength			length of this descriptor = 9 bytes
                        ;bDescriptor type	CONFIGURATION descriptor = 2
.DB     67,0 			;wTotalLength		total length of configuration block = 67 bytes
                        
.DB     2,1      		;bNumInterfaces		two interfaces incl this descriptor
						;bConfigurationValue ID of this configuration = 1
#ifdef SELFPOWERED
.DB     0,0x40          ;iConfiguration		index of string descriptor:unused=0
						;bmAttributes		SELF-powered - no remote wakeup
.DB     0,9             ;MaxPower 			= 0 (self-powered only)
						;bLength (of next)	length of this descriptor = 9 bytes
#else
.DB     0,0x80          ;iConfiguration		index of string descriptor:unused=0
						;bmAttributes		BUS-powered - no remote wakeup
.DB     49,9            ;MaxPower			entry*2 = 98mA
						;bLength of next descriptor = 9 bytes
//9 belongs to next block.Put here due to HalfWord configuration
#endif
;COMMUNICATION DEVICE CLASS standard INTERFACE Descriptor  
;length of this descriptor = 9 bytes
.DB     4,0          	;bDescriptorType	INTERFACE descriptor type=4 
                        ;bInterfaceNumber	index of this interface =0 
.DB     0,1		        ;bAlternateSetting	index of this setting =0 
                        ;bNumEndpoints 		(Notification endpoint) = 1 
.DB     2,2       		;bInterfaceClass 	= 2 (COMMUNICATION)
                        ;bInterfaceSubClass = 2 (ABSTRACT_CONTROL_MODEL)
.DB     0,0         	;bInterfaceProtocol = 0 (unused)
                        ;iInterface 		= 0 (unused)
;COMMUNICATION DEVICE specific functional descriptor
.DB     5,0x24       	;bLength			length of this descriptor = 5 bytes
                        ;bDescriptorType	CS_INTERFACEdescriptor type = 0x24
.DB     0,0x10       	;bDescriptortSubtype HEADER subtype
.DB     1,5             ;bcdADC				=0x0100 revision of class specific descriptors
						;bLength of next descriptor = 5 bytes
;CALL MANAGEMENT functional Descriptor
;length of this descriptor = 5 bytes
.DB     0x24,1 			;bDescriptorType	CS_INTERFACEdescriptor type = 0x24
						;bDescriptorSubtype	
.DB     3,0   			;bmCapabilities		Device handles call management by itself 
						;					ia DataInterface
						;					i.e. in this case: no call management required
						;bDataInterface:	Data interface no.0 (see below) associated
;ABSTRACT CONTROL management functional Descriptor 
.DB     4,0x24         	;bLength			length of this descriptor = 4 bytes
						;bDesriptorType		CS_INTERFACEdescriptor type = 0x24
.DB     2,7				;bDescriptorSubtype
						;bmCapabilities		:Send_Break,Set_Line_Coding... etc
;UNION functional descriptor
.DB 	5,0x24          ;bLength			length of this descriptor = 5 bytes
                        ;bDescriptorType	CS_INTERFACE descriptor
.DB 	6,0		       	;bDescriptorSubtype UNION FUNCTIONAL descriptor
                        ;bMasterInterface
.DB     1,7  			;bSlaveInterface0    
						;bLength of next descriptor = 7 bytes    
;Notification standard endpoint descriptor
.DB		5,0x83			;bDescriptorType	ENDPOINT
						;bEndpointAddress	n0.3, input
.DB		3,16			;bmAttributes 		INTERRUPT = 3
						;wMaxPacketSize		= 16
.DB		0,64			;interval in ms		= 64
			            
;DATA CLASS INTERFACE descriptor
.DB		9,4				;bLength			length of this descriptor = 9 bytes
						;bDescriptorType	INTERFACE descriptor type
.DB		1,0				;bInterfaceNumber
						;bAlternateSetting
.DB		2,0x0A			;bNumEndpoints
						;bInterfaceClass	Data Class
.DB		0,0				;bInterfaceSubclass unused
						;bInterfaceProtocol no specific protocol
.DB		0,7				;iInterface 		unused
						;bLength of next descriptor = 7 bytes
;DATA BULK IN endpoint standard descriptor
.DB		5,0x81			;bDescriptorType	ENDPOINT
						;bEndpointAddress	nr.1, input 
.DB		2,64			;bmAttributes		= 2 (bulk)
						;wMaxPacketSize		= 64
.DB		0,64			;bInterval			=64ms  ;special for DMX

;DATA BULK OUT endpoint standard descriptor
.DB		7,5				;bLength			length of this descriptor = 7 bytes
						;bDescriptorType	ENDPOINT 
.DB		2,2				;bEndpointAddress	nr.=2, output
						;bmAttributes 		= 2 (bulk)
.DB		64,0			;wMaxPacketSize		= 64

.DB		0,10			;bInterval			=10ms  ;special for DMX
						;padding byte

//CONFIG_DESCRIPTOR:

MANUFACT_STRINGDESCRIPTOR:;bString (Unicode 16Bit)
;Product String Deskriptor index 1
                                                ;bString (Unicode 16Bit)
.DB     18,0x03         ;bLength = 16+2 Groesse Deskriptor in Byte
                        ;bDescriptorType = 0x03, Konstante String = 3
.DB     'h',0,'o',0,'m',0,'e',0,'m',0,'a',0,'d',0,'e',0


PRODUCT_STRINGDESCRIPTOR:;bString (Unicode 16Bit)
;Product String Deskriptor index 1
                                                ;bString (Unicode 16Bit)
.DB     18,0x03         ;bLength = 26+2 Groesse Deskriptor in Byte
                        ;bDescriptorType = 0x03, Konstante String = 3
.DB     'A',0,'v',0,'i',0,'s',0,'e',0,'-',0,'5',0,'6',0

FAKELINECODING:
.DB		0x00,0xC2,0x01,0x00,0,0,8,0  ;0x0001C200=decimal 115200, rest describes 1N8 + paddingbyte
;the baud rate is faked, the microcontroller accepts any and sets its own at USART.
;but 115200 seems to be the best default at host side. Else enter your baudrate 32bit formatted.

;NULL_NOTIFICATION:  ;SERIAL_STATE response seems not to be mandatory
;.DB 	0xA1,0x20,0,0,0,0,2,0,0,0

DEVICE_STATUS: 
.DW 0					//Bus powered, no Remote Wakep



